# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Test of the Verilog parser
"""

from unittest import TestCase
import os
from os.path import join, dirname, exists
import time
import shutil
from vunit.ostools import renew_path
from vunit.parsing.verilog.parser import VerilogParser
from vunit.test.mock_2or3 import mock


class TestVerilogParser(TestCase):  # pylint: disable=too-many-public-methods
    """
    Test of the Verilog parser
    """

    def setUp(self):
        self.output_path = join(dirname(__file__), "test_verilog_parser_out")
        renew_path(self.output_path)
        self.cwd = os.getcwd()
        os.chdir(self.output_path)

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(self.output_path)

    def test_parsing_empty(self):
        design_file = self.parse("")
        self.assertEqual(design_file.modules, [])

    def test_parse_module(self):
        modules = self.parse("""\
module true1;
  my_module hello
  "module false";
endmodule
/*
module false
*/
  module true2; // module false
endmodule module true3
endmodule
""").modules
        self.assertEqual(len(modules), 3)
        self.assertEqual(modules[0].name, "true1")
        self.assertEqual(modules[1].name, "true2")
        self.assertEqual(modules[2].name, "true3")

    def test_parse_parameter_without_type(self):
        modules = self.parse("""\
module foo;
  parameter param1;
  parameter param2 = 1;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_parse_parameter_with_type(self):
        modules = self.parse("""\
module foo;
  parameter string param1;
  parameter integer param2 = 1;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_nested_modules_are_ignored(self):
        modules = self.parse("""\
module foo;
  parameter string param1;
  module nested;
    parameter integer param_nested;
  endmodule
  parameter string param2;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_parse_package(self):
        packages = self.parse("""\
package true1;
endpackage package true2; endpackage
""").packages
        self.assertEqual(len(packages), 2)
        self.assertEqual(packages[0].name, "true1")
        self.assertEqual(packages[1].name, "true2")

    def test_parse_imports(self):
        imports = self.parse("""\
import true1;
package pkg;
  import true2::*;
endpackage
""").imports
        self.assertEqual(len(imports), 2)
        self.assertEqual(imports[0], "true1")
        self.assertEqual(imports[1], "true2")

    def test_parse_package_references(self):
        package_references = self.parse("""\
import false1;
import false1::false2::*;
package pkg;
  true1::func(true2::bar());
  true3::foo();
endpackage
""").package_references
        self.assertEqual(len(package_references), 3)
        self.assertEqual(package_references[0], "true1")
        self.assertEqual(package_references[1], "true2")
        self.assertEqual(package_references[2], "true3")

    @mock.patch("vunit.parsing.verilog.parser.LOGGER", autospec=True)
    def test_parse_import_with_bad_argument(self, logger):
        imports = self.parse("""\
import;
""").imports
        self.assertEqual(len(imports), 0)
        logger.warning.assert_called_once_with(
            "import bad argument\n%s",
            "at file_name.sv line 1:\n"
            "import;\n"
            "      ~")

    @mock.patch("vunit.parsing.verilog.parser.LOGGER", autospec=True)
    def test_parse_import_eof(self, logger):
        imports = self.parse("""\
import
""").imports
        self.assertEqual(len(imports), 0)
        logger.warning.assert_called_once_with(
            "EOF reached when parsing import\n%s",
            "at file_name.sv line 1:\n"
            "import\n"
            "~~~~~~")

    def test_parse_instances(self):
        instances = self.parse("""\
module name;
  true1 instance_name1();
  true2 instance_name2(.foo(bar));
  true3 #(.param(1)) instance_name3(.foo(bar));
endmodule
""").instances
        self.assertEqual(len(instances), 3)
        self.assertEqual(instances[0], "true1")
        self.assertEqual(instances[1], "true2")
        self.assertEqual(instances[2], "true3")

    def test_parse_instances_after_block_label(self):
        instances = self.parse("""\
module name;
genvar i;
  generate
    for( i=0; i < 10; i = i + 1 )
      begin: INST_GEN
        true1 instance_name1();
    end : INST_GEN
    true2 instance_name2();
  endgenerate
endmodule
""").instances
        self.assertEqual(len(instances), 2)
        self.assertEqual(instances[0], "true1")
        self.assertEqual(instances[1], "true2")

    def test_parse_instances_without_crashing(self):
        instances = self.parse("""\
module name;
endmodule identifier
""").instances
        self.assertEqual(len(instances), 0)

    def test_can_set_pre_defined_defines(self):
        code = """\
`ifdef foo
`foo
endmodule;
`endif
"""

        result = self.parse(code, defines={"foo": "module mod1;"})
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod1")

    def test_result_is_cached(self):
        code = """\
`include "missing.sv"
module name;
  true1 instance_name1();
  true2 instance_name2(.foo(bar));
  true3 #(.param(1)) instance_name3(.foo(bar));
endmodule
"""
        cache = {}
        result = self.parse(code, cache=cache)
        instances = result.instances
        self.assertEqual(len(instances), 3)
        self.assertEqual(instances[0], "true1")
        self.assertEqual(instances[1], "true2")
        self.assertEqual(instances[2], "true3")

        new_result = self.parse(code, cache=cache)
        self.assertEqual(id(result), id(new_result))
        cache.clear()

        new_result = self.parse(code, cache=cache)
        self.assertNotEqual(id(result), id(new_result))

    def test_cached_parsing_updated_by_changing_file(self):
        code = """\
module mod1;
endmodule
"""
        cache = {}
        result = self.parse(code, cache=cache)
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod1")

        tick()

        code = """\
module mod2;
endmodule
"""
        result = self.parse(code, cache=cache)
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod2")

    def test_cached_parsing_updated_by_includes(self):
        self.write_file("include.svh", """
module mod;
endmodule;
""")
        code = """\
`include "include.svh"
"""
        cache = {}
        result = self.parse(code, cache=cache, include_paths=[self.output_path])
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod")

        tick()

        self.write_file("include.svh", """
module mod1;
endmodule;

module mod2;
endmodule;
""")
        result = self.parse(code, cache=cache, include_paths=[self.output_path])
        self.assertEqual(len(result.modules), 2)
        self.assertEqual(result.modules[0].name, "mod1")
        self.assertEqual(result.modules[1].name, "mod2")

    def test_cached_parsing_updated_by_higher_priority_file(self):
        cache = {}
        include_paths = [self.output_path, join(self.output_path, "lower_prio")]

        self.write_file(join("lower_prio", "include.svh"), """
module mod_lower_prio;
endmodule;
""")

        code = """\
`include "include.svh"
"""

        result = self.parse(code, cache=cache, include_paths=include_paths)
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod_lower_prio")

        self.write_file("include.svh", """
module mod_higher_prio;
endmodule;
""")
        result = self.parse(code, cache=cache, include_paths=include_paths)
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod_higher_prio")

    def test_cached_parsing_updated_by_other_defines(self):
        cache = {}

        code = """\
`ifdef foo
module `foo
endmodule;
`endif
"""

        result = self.parse(code, cache=cache)
        self.assertEqual(len(result.modules), 0)

        result = self.parse(code, cache=cache, defines={"foo": "mod1"})
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod1")

        result = self.parse(code, cache=cache, defines={"foo": "mod2"})
        self.assertEqual(len(result.modules), 1)
        self.assertEqual(result.modules[0].name, "mod2")

    def write_file(self, file_name, contents):
        """
        Write file with contents into output path
        """
        full_name = join(self.output_path, file_name)
        full_path = dirname(full_name)
        if not exists(full_path):
            os.makedirs(full_path)
        with open(full_name, "w") as fptr:
            fptr.write(contents)

    def parse(self, code, include_paths=None, cache=None, defines=None):
        """
        Helper function to parse
        """
        self.write_file("file_name.sv", code)
        cache = cache if cache is not None else {}
        parser = VerilogParser(database=cache)
        include_paths = include_paths if include_paths is not None else []
        design_file = parser.parse("file_name.sv", include_paths, defines)
        return design_file


def tick():
    """
    To get a different file modification time
    """
    time.sleep(0.01)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the DependencyGraph
"""

import unittest
from vunit.dependency_graph import (DependencyGraph,
                                    CircularDependencyException)


class TestDependencyGraph(unittest.TestCase):
    """
    Test the DependencyGraph
    """

    def test_should_return_empty_compile_order_for_no_nodes(self):
        graph = DependencyGraph()
        self.assertEqual(graph.toposort(), [], 'Should return empty list')

    def test_should_return_list_of_nodes_when_there_are_no_dependencies(self):
        nodes = ['a', 'b', 'c', 'd']
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, [])
        result = graph.toposort()
        self.assertEqual(result.sort(), nodes.sort(), 'Should return the node list in any order')

    def test_should_sort_in_topological_order_when_there_are_dependencies(self):
        nodes = ['a', 'b', 'c', 'd', 'e', 'f']
        dependencies = [('a', 'b'), ('a', 'c'), ('b', 'd'), ('e', 'f')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.toposort()
        self._check_result(result, dependencies)

    def test_should_raise_runtime_error_exception_on_self_dependency(self):
        nodes = ['a', 'b', 'c', 'd']
        dependencies = [('a', 'b'), ('a', 'c'), ('b', 'd'), ('d', 'd')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.toposort()
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ['d', 'd'])
        else:
            self.fail("Exception not raised")

    def test_should_raise_runtime_error_exception_on_long_circular_dependency(self):
        nodes = ['a', 'b', 'c', 'd']
        dependencies = [('a', 'b'), ('a', 'c'), ('b', 'd'), ('d', 'a')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.toposort()
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ['a', 'b', 'd', 'a'])
        else:
            self.fail("Exception not raised")

    def test_should_resort_after_additions(self):
        nodes = ['a', 'b', 'c', 'd', 'e', 'f']
        dependencies = [('a', 'b'), ('a', 'c'), ('b', 'd'), ('e', 'f')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        graph.toposort()
        dependencies = [('a', 'b'), ('a', 'c'), ('b', 'd'), ('e', 'f'), ('b', 'g')]
        graph.add_node('g')
        graph.add_dependency('b', 'g')
        result = graph.toposort()
        self._check_result(result, dependencies)

    def test_get_direct_dependencies_should_return_empty_set_when_no_dependendencies(self):
        nodes = ['a', 'b', 'c']
        dependencies = []
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.get_direct_dependencies('b')
        self.assertTrue(isinstance(result, (set)))
        self.assertFalse(result)

    def test_get_direct_dependencies_should_return_dependendencies_set(self):
        nodes = ['a', 'b', 'c', 'd']
        dependencies = [('a', 'b'), ('a', 'c')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.get_direct_dependencies('c')
        self.assertFalse('b' in result)
        self.assertTrue('a' in result)

    def test_get_dependent(self):
        nodes = ['a', 'b', 'c', 'd', 'e']
        dependencies = [('a', 'b'), ('b', 'c'), ('c', 'd'), ('e', 'd')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        self.assertEqual(graph.get_dependent(set('a')), set(('a', 'b', 'c', 'd')))
        self.assertEqual(graph.get_dependent(set('e')), set(('d', 'e')))

    def test_get_dependencies(self):
        nodes = ['a', 'b', 'c', 'd', 'e']
        dependencies = [('b', 'a'), ('c', 'b'), ('d', 'c'), ('d', 'e')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        self.assertEqual(graph.get_dependencies(set('a')), set(('a', 'b', 'c', 'd')))
        self.assertEqual(graph.get_dependencies(set('e')), set(('d', 'e')))

    def test_get_dependent_detects_circular_dependencies(self):
        nodes = ['a', 'b', 'c']
        dependencies = [('a', 'b'), ('b', 'c'), ('c', 'a')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.get_dependent(set('a'))
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ['a', 'b', 'c', 'a'])
        else:
            self.fail("Exception not raised")

    def test_get_dependencies_detects_circular_dependencies(self):
        nodes = ['a', 'b', 'c']
        dependencies = [('b', 'a'), ('c', 'b'), ('a', 'c')]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.get_dependencies(set('a'))
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ['a', 'b', 'c', 'a'])
        else:
            self.fail("Exception not raised")

    def _check_result(self, result, dependencies):
        """
        Check that the resulting has an order such that
        dependent files are located after their dependencies
        """
        for dep1, dep2 in dependencies:
            self.assertTrue(result.index(dep1) < result.index(dep2), "%s is not before %s" % (dep1, dep2))

    @staticmethod
    def _add_nodes_and_dependencies(graph, nodes, dependencies):
        """
        Add nodes and dependencies to the graph
        """
        for node in nodes:
            graph.add_node(node)
        for dep in dependencies:
            graph.add_dependency(dep[0], dep[1])

if __name__ == '__main__':
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Acceptance test of VUnit end to end functionality
"""


import unittest
from os.path import join, dirname
from os import environ
from subprocess import call
import sys
from vunit.test.common import has_simulator, check_report, simulator_is


@unittest.skipUnless(has_simulator(), "Requires simulator")
class TestVunitArtificial(unittest.TestCase):
    """
    Acceptance test of VUnit end to end functionality using
    artificial test benches.
    """
    def setUp(self):
        # Spaces in path intentional to verify that it is supported
        self.output_path = join(dirname(__file__), "artificial out")
        self.report_file = join(self.output_path, "xunit.xml")
        self.artificial_run_vhdl = join(dirname(__file__), "artificial", "vhdl", "run.py")
        self.artificial_run_verilog = join(dirname(__file__), "artificial", "verilog", "run.py")

    @unittest.skipUnless(simulator_is("modelsim", "rivierapro"),
                         "Only simulators with persistance functionality")
    def test_artificial_modelsim_unique_sim(self):
        self._test_artificial(args=["--unique-sim"])

    def test_artificial(self):
        self._test_artificial()

    def test_artificial_elaborate_only(self):
        self.check(self.artificial_run_vhdl,
                   exit_code=1,
                   args=["--elaborate"])

        elab_expected_report = []
        for status, name in EXPECTED_REPORT:
            if name in ("lib.tb_elab_fail.all",):
                status = "failed"
            else:
                status = "passed"
            elab_expected_report.append((status, name))

        check_report(self.report_file, elab_expected_report)

        self.check(self.artificial_run_vhdl,
                   exit_code=0,
                   clean=False,
                   args=["--elaborate", "lib.tb_pass.all"])
        check_report(self.report_file, [
            ("passed", "lib.tb_pass.all")])

        self.check(self.artificial_run_vhdl,
                   exit_code=1,
                   clean=False,
                   args=["--elaborate", "lib.tb_elab_fail.all"])
        check_report(self.report_file, [
            ("failed", "lib.tb_elab_fail.all")])

    def _test_artificial(self, args=None):
        """
        Utility function to run and check the result of all test benches
        using either persistent or non-persistent simulator interface mode
        """
        self.check(self.artificial_run_vhdl,
                   exit_code=1,
                   args=args)
        check_report(self.report_file, EXPECTED_REPORT)

    def test_run_selected_tests_in_same_sim_test_bench_vhdl(self):
        self._test_run_selected_tests_in_same_sim_test_bench(self.artificial_run_vhdl)

    @unittest.skipUnless(simulator_is("modelsim"), "Only modelsim supports verilog")
    def test_run_selected_tests_in_same_sim_test_bench_verilog(self):
        self._test_run_selected_tests_in_same_sim_test_bench(self.artificial_run_verilog)

    def _test_run_selected_tests_in_same_sim_test_bench(self, run_file):
        """
        Run selected "same_sim" test in isolation
        """
        self.check(run_file,
                   exit_code=0,
                   clean=True,
                   args=["*same_sim_some_fail*Test 1*"])
        check_report(self.report_file, [
            ("passed", "lib.tb_same_sim_some_fail.Test 1")])

        self.check(run_file,
                   exit_code=1,
                   clean=False,
                   args=["*same_sim_some_fail*Test 2*"])
        check_report(self.report_file, [
            ("failed", "lib.tb_same_sim_some_fail.Test 2")])

        self.check(run_file,
                   exit_code=0,
                   clean=False,
                   args=["*same_sim_some_fail*Test 3*"])
        check_report(self.report_file, [
            ("passed", "lib.tb_same_sim_some_fail.Test 3")])

        self.check(run_file,
                   exit_code=1,
                   clean=False,
                   args=["*same_sim_some_fail*Test 2*", "*same_sim_some_fail*Test 3*"])
        check_report(self.report_file, [
            ("failed", "lib.tb_same_sim_some_fail.Test 2"),
            ("skipped", "lib.tb_same_sim_some_fail.Test 3")])

    @unittest.skipUnless(simulator_is("modelsim"), "Only modelsim supports verilog")
    def test_artificial_verilog(self):
        self.check(self.artificial_run_verilog,
                   exit_code=1)
        check_report(self.report_file, [
            ("passed", "lib.tb_other_file_tests.pass"),
            ("failed", "lib.tb_other_file_tests.fail"),

            ("passed", "lib.tb_magic_paths.Test magic paths are correct"),
            ("passed", "lib.tb_with_define.test 1"),

            ("failed", "lib.tb_fail_on_warning.fail"),

            ("failed", "lib.tb_fail_on_fatal_and_early_finish.fatal0"),
            ("failed", "lib.tb_fail_on_fatal_and_early_finish.fatal1"),
            ("failed", "lib.tb_fail_on_fatal_and_early_finish.finish0"),
            ("failed", "lib.tb_fail_on_fatal_and_early_finish.finish1"),

            ("passed", "lib.tb_with_parameter_config.Test 0"),
            ("passed", "lib.tb_with_parameter_config.cfg.Test 1"),
            ("passed", "lib.tb_with_parameter_config.Test 2"),
            ("passed", "lib.tb_with_parameter_config.cfg.Test 3"),
            ("passed", "lib.tb_with_parameter_config.cfg.Test 4"),

            ("passed", "lib.tb_same_sim_all_pass.cfg.Test 1"),
            ("passed", "lib.tb_same_sim_all_pass.cfg.Test 2"),
            ("passed", "lib.tb_same_sim_all_pass.cfg.Test 3"),

            ("passed", "lib.tb_same_sim_some_fail.Test 1"),
            ("failed", "lib.tb_same_sim_some_fail.Test 2"),
            ("skipped", "lib.tb_same_sim_some_fail.Test 3"),

            ("passed", "lib.tb_with_runner.pass"),
            ("failed", "lib.tb_with_runner.fail")])

    # pylint: disable=too-many-arguments
    def check(self, run_file, args=None, clean=True, exit_code=0):
        """
        Run external run file and verify exit code
        """
        args = args if args is not None else []
        new_env = environ.copy()
        new_env["VUNIT_VHDL_STANDARD"] = '2008'
        if clean:
            args += ["--clean"]
        retcode = call([sys.executable, run_file,
                        "--output-path=%s" % self.output_path,
                        "--xunit-xml=%s" % self.report_file] + args,
                       env=new_env)
        self.assertEqual(retcode, exit_code)

    def test_exit_0_flag(self):
        self.check(self.artificial_run_vhdl,
                   exit_code=1,
                   args=["lib.tb_fail.all"])
        self.check(self.artificial_run_vhdl,
                   exit_code=0,
                   args=["--exit-0", "lib.tb_fail.all"])


EXPECTED_REPORT = (
    ("passed", "lib.tb_other_file_tests.pass"),
    ("failed", "lib.tb_other_file_tests.fail"),
    ("passed", "lib.tb_pass.all"),
    ("failed", "lib.tb_fail.all"),
    ("passed", "lib.tb_infinite_events.all"),
    ("failed", "lib.tb_fail_on_warning.all"),
    ("passed", "lib.tb_no_fail_on_warning.all"),
    ("passed", "lib.tb_with_vhdl_runner.pass"),
    ("passed", "lib.tb_with_vhdl_runner.Test with spaces"),
    ("failed", "lib.tb_with_vhdl_runner.fail"),
    ("failed", "lib.tb_with_vhdl_runner.Test that timeouts"),
    ("passed", "lib.tb_magic_paths.all"),
    ("passed", "lib.tb_no_fail_after_cleanup.all"),
    ("failed", "lib.tb_elab_fail.all"),

    ("passed", "lib.tb_same_sim_all_pass.cfg.Test 1"),
    ("passed", "lib.tb_same_sim_all_pass.cfg.Test 2"),
    ("passed", "lib.tb_same_sim_all_pass.cfg.Test 3"),

    ("passed", "lib.tb_same_sim_some_fail.Test 1"),
    ("failed", "lib.tb_same_sim_some_fail.Test 2"),
    ("skipped", "lib.tb_same_sim_some_fail.Test 3"),

    ("passed", "lib.tb_with_checks.Test passing check"),
    ("failed", "lib.tb_with_checks.Test failing check"),
    ("failed", "lib.tb_with_checks.Test non-stopping failing check"),

    ("passed", "lib.tb_set_generic.all"),

    ("passed", "lib.tb_with_generic_config.Test 0"),
    ("passed", "lib.tb_with_generic_config.cfg.Test 1"),
    ("passed", "lib.tb_with_generic_config.Test 2"),
    ("passed", "lib.tb_with_generic_config.cfg.Test 3"),
    ("passed", "lib.tb_with_generic_config.cfg.Test 4"),

    ("passed", "lib.tb_no_generic_override.all"),

    ("passed", "lib.tb_ieee_warning.pass"),
    ("failed", "lib.tb_ieee_warning.fail"),

    ("failed", "lib.tb_assert_stop_level.Report warning when VHDL assert stop level = warning"),
    ("failed", "lib.tb_assert_stop_level.Report error when VHDL assert stop level = warning"),
    ("failed", "lib.tb_assert_stop_level.Report failure when VHDL assert stop level = warning"),
    ("passed", "lib.tb_assert_stop_level.Report warning when VHDL assert stop level = error"),
    ("failed", "lib.tb_assert_stop_level.Report error when VHDL assert stop level = error"),
    ("failed", "lib.tb_assert_stop_level.Report failure when VHDL assert stop level = error"),
    ("passed", "lib.tb_assert_stop_level.Report warning when VHDL assert stop level = failure"),
    ("passed", "lib.tb_assert_stop_level.Report error when VHDL assert stop level = failure"),
    ("failed", "lib.tb_assert_stop_level.Report failure when VHDL assert stop level = failure"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Adding Custom Command Line Arguments
------------------------------------
It is possible to add custom command line arguments to your ``run.py``
scripts using the :class:`.VUnitCLI` class.

.. autoclass:: vunit.vunit_cli.VUnitCLI
   :members:

A :class:`.VUnitCLI` object has a ``parser`` field which is an
`ArgumentParser` object of the `argparse`_ library.

.. _argparse: https://docs.python.org/3/library/argparse.html

.. code-block:: python

   from vunit import VUnitCLI, VUnit

   # Add custom command line argument to standard CLI
   # Beware of conflicts with existing arguments
   cli = VUnitCLI()
   cli.parser.add_argument('--custom-arg', ...)
   args = cli.parse_args()

   # Create VUNit instance from custom arguments
   vu = VUnit.from_args(args=args)

   # Use args.custom_arg here ...
   print(args.custom_arg)

"""

import argparse
from os.path import join, abspath
import os
from vunit.simulator_factory import SIMULATOR_FACTORY
from vunit.about import version


class VUnitCLI(object):
    """
    VUnit command line interface
    """

    def __init__(self, description=None):
        """
        :param description: A custom short description of the command line tool
        """
        self.parser = _create_argument_parser(description)

    def parse_args(self, argv=None):
        """
        Parse command line arguments

        :param argv: Use explicit argv instead of actual command line argument
        :returns: The parsed argument namespace object
        """
        return self.parser.parse_args(args=argv)


def _create_argument_parser(description=None, for_documentation=False):
    """
    Create the argument parser

    :param description: A custom short description of the command line tool
    :param for_documentation: When used for user guide documentation
    :returns: The created :mod:`argparse` parser object
    """
    if description is None:
        description = 'VUnit command line tool version %s' % version()

    if for_documentation:
        default_output_path = "./vunit_out"
    else:
        default_output_path = join(abspath(os.getcwd()), "vunit_out")

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('test_patterns', metavar='tests', nargs='*',
                        default='*',
                        help='Tests to run')

    parser.add_argument('-l', '--list', action='store_true',
                        default=False,
                        help='Only list all test cases')

    parser.add_argument('-f', '--files', action='store_true',
                        default=False,
                        help='Only list all files in compile order')

    parser.add_argument('--compile', action='store_true',
                        default=False,
                        help='Only compile project without running tests')

    parser.add_argument('-k', '--keep-compiling', action='store_true',
                        default=False,
                        help='Continue compiling even after errors only skipping files that depend on failed files')

    parser.add_argument('--elaborate', action='store_true',
                        default=False,
                        help='Only elaborate test benches without running')

    parser.add_argument('--clean', action='store_true',
                        default=False,
                        help='Remove output path first')

    parser.add_argument('-o', '--output-path',
                        default=default_output_path,
                        help='Output path for compilation and simulation artifacts')

    parser.add_argument('-x', '--xunit-xml',
                        default=None,
                        help='Xunit test report .xml file')

    parser.add_argument('--exit-0',
                        default=False,
                        action="store_true",
                        help=('Exit with code 0 even if a test failed. '
                              'Still exits with code 1 on fatal errors such as compilation failure'))

    parser.add_argument('--dont-catch-exceptions',
                        default=False,
                        action="store_true",
                        help=('Let exceptions bubble up all the way. '
                              'Useful when running with "python -m pdb".'))

    parser.add_argument('-v', '--verbose', action="store_true",
                        default=False,
                        help='Print test output immediately and not only when failure')

    parser.add_argument('-q', '--quiet', action="store_true",
                        default=False,
                        help='Do not print test output even in the case of failure')

    parser.add_argument('--no-color', action='store_true',
                        default=False,
                        help='Do not color output')

    parser.add_argument('--log-level',
                        default="warning",
                        choices=["info", "error", "warning", "debug"],
                        help=("Log level of VUnit internal python logging. "
                              "Used for debugging"))

    parser.add_argument('-p', '--num-threads', type=positive_int,
                        default=1,
                        help=('Number of tests to run in parallel. '
                              'Test output is not continuously written in verbose mode with p > 1'))

    parser.add_argument("-u", "--unique-sim",
                        action="store_true",
                        default=False,
                        help="Do not re-use the same simulator process for running different test cases (slower)")

    parser.add_argument("--coverage",
                        default=None,
                        nargs="?",
                        help="Enable code coverage. Works with ModelSim and RivieraPRO.")

    parser.add_argument('--version', action='version', version=version())

    SIMULATOR_FACTORY.add_arguments(parser,
                                    for_all_simulators=for_documentation)

    com = parser.add_argument_group("com", description="Flags specific to the com message passing package")
    com.add_argument('--use-debug-codecs', action='store_true',
                     default=False,
                     help='Run with debug features enabled')

    return parser


def positive_int(val):
    """
    ArgumentParse positive int check
    """
    try:
        ival = int(val)
        assert ival > 0
        return ival
    except (ValueError, AssertionError):
        raise argparse.ArgumentTypeError("'%s' is not a valid positive int" % val)


def _parser_for_documentation():
    """
    Returns an argparse object used by sphinx for documentation in user_guide.rst
    """
    return _create_argument_parser(for_documentation=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Preprocessing of VHDL files to add file_name and line_num arguments to function calls
to enable better messages
"""


import re


class LocationPreprocessor(object):
    """
    Preprocessing of VHDL files to add file_name and line_num
    arguments to calls of known function and procedures
    """
    def __init__(self):
        self._subprograms_with_arguments = [
            'log', 'verbose_high2', 'verbose_high1', 'verbose',
            'verbose_low1', 'verbose_low2', 'debug_high2', 'debug_high1',
            'debug', 'debug_low1', 'debug_low2', 'info_high2', 'info_high1',
            'info', 'info_low1', 'info_low2', 'warning_high2', 'warning_high1',
            'warning', 'warning_low1', 'warning_low2', 'error_high2',
            'error_high1', 'error', 'error_low1', 'error_low2', 'failure_high2',
            'failure_high1', 'failure', 'failure_low1', 'failure_low2', 'check',
            'check_failed', 'check_true', 'check_false', 'check_implication',
            'check_stable', 'check_equal', 'check_not_unknown', 'check_zero_one_hot',
            'check_one_hot', 'check_next', 'check_sequence', 'check_relation',
            'lock_entry', 'lock_exit', 'unlock_entry', 'unlock_exit']
        self._subprograms_without_arguments = []

    def add_subprogram(self, subprogram):
        """
        Add a subprogram name to the list of known names to preprocess
        """
        self._subprograms_without_arguments.append(subprogram)
        self._subprograms_with_arguments.append(subprogram)

    def remove_subprogram(self, subprogram):
        """
        Remove a subprogram name from the list of known names to preprocess
        """
        if subprogram not in self._subprograms_without_arguments + self._subprograms_with_arguments:
            raise RuntimeError('Unable to remove unknown subprogram %s' % subprogram)

        if subprogram in self._subprograms_without_arguments:
            self._subprograms_without_arguments.remove(subprogram)

        if subprogram in self._subprograms_with_arguments:
            self._subprograms_with_arguments.remove(subprogram)

    @staticmethod
    def _find_closing_parenthesis(args):
        """
        Find the balanced closing parentesis

        @TODO duplicate with vhdl_parser.py
        """
        pattern = re.compile(r'\(|\)')
        balance = 0
        for match in pattern.finditer(args):
            balance += 1 if match.group() == '(' else -1
            if balance == 0:
                return match.start()
        return None

    _already_fixed_file_name_pattern = re.compile(r'file_name\s*=>', re.MULTILINE)
    _already_fixed_line_num_pattern = re.compile(r'line_num\s*=>', re.MULTILINE)
    _subprogram_declaration_start_backwards_pattern = re.compile(r'\s+(erudecorp|noitcnuf)')
    _assignment_pattern = re.compile(r'\s*(:=|<=)', re.MULTILINE)

    def run(self, code, file_name):
        """
        Return preprocessed code given file_name of original file
        """
        potential_subprogram_call_with_arguments_pattern = re.compile(
            r'[^a-zA-Z0-9_](?P<subprogram>' + '|'.join(self._subprograms_with_arguments) + r')\s*(?P<args>\()',
            re.MULTILINE)

        potential_subprogram_call_without_arguments_pattern = re.compile(
            r'[^a-zA-Z0-9_](?P<subprogram>' + '|'.join(self._subprograms_without_arguments) + r')\s*;',
            re.MULTILINE)

        matches = list(potential_subprogram_call_with_arguments_pattern.finditer(code))
        if self._subprograms_without_arguments:
            matches += list(potential_subprogram_call_without_arguments_pattern.finditer(code))
        matches.sort(key=lambda match: match.start('subprogram'), reverse=True)

        for match in matches:
            if self._subprogram_declaration_start_backwards_pattern.match(code[match.start():0:-1]):
                continue
            file_name_association = ', file_name => "' + file_name + '"'
            line_num_association = ', line_num => ' + str(1 + code[:match.start('subprogram')].count('\n'))
            if 'args' in match.groupdict():
                closing_paranthesis_start = self._find_closing_parenthesis(code[match.start('args'):])

                if self._assignment_pattern.match(code[match.start('args') + closing_paranthesis_start + 1:]):
                    continue

                args = code[match.start('args'): match.start('args') + closing_paranthesis_start]
                already_fixed_file_name = self._already_fixed_file_name_pattern.search(args) is not None
                already_fixed_line_num = self._already_fixed_line_num_pattern.search(args) is not None
                file_name_association = file_name_association if not already_fixed_file_name else ''
                line_num_association = line_num_association if not already_fixed_line_num else ''

                code = (code[:match.start('args') + closing_paranthesis_start] +
                        line_num_association +
                        file_name_association + code[match.start('args') + closing_paranthesis_start:])
            else:
                code = (code[:match.end('subprogram')] +
                        '(' + line_num_association[2:] + file_name_association + ')' +
                        code[match.end('subprogram'):])
        return code



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2016, Lars Asplund lars.anders.asplund@gmail.com

# pylint: skip-file
from sys import stdout

if hasattr(stdout, "buffer"):
    # Python 3
    stdout.buffer.write(b'a' + b"\x87" + b'c')
else:
    # Python 2.7
    stdout.write(b'a' + b"\x87" + b'c')

stdout.write("\n")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=unused-wildcard-import
# pylint: disable=wildcard-import

"""
Test of the Verilog tokenizer
"""

from unittest import TestCase
from vunit.parsing.verilog.tokenizer import VerilogTokenizer
from vunit.parsing.verilog.tokens import *


class TestVerilogTokenizer(TestCase):
    """
    Test of the Verilog tokenizer
    """

    def test_tokenizes_define(self):
        self.check("`define name",
                   [PREPROCESSOR(value="define"),
                    WHITESPACE(value=" "),
                    IDENTIFIER(value="name")])

    def test_tokenizes_string_literal(self):
        self.check('"hello"',
                   [STRING(value='hello')])

        self.check('"hel""lo"',
                   [STRING(value='hel'),
                    STRING(value='lo')])

        self.check(r'"h\"ello"',
                   [STRING(value='h"ello')])

        self.check(r'"h\"ello"',
                   [STRING(value='h"ello')])

        self.check(r'"\"ello"',
                   [STRING(value='"ello')])

        self.check(r'"\"\""',
                   [STRING(value='""')])

        self.check(r'''"hi
there"''',
                   [STRING(value='hi\nthere')])

        self.check(r'''"hi\
there"''',
                   [STRING(value='hithere')])

    def test_tokenizes_single_line_comment(self):
        self.check("// asd",
                   [COMMENT(value=" asd")])

        self.check("asd// asd",
                   [IDENTIFIER(value="asd"),
                    COMMENT(value=" asd")])

        self.check("asd// asd //",
                   [IDENTIFIER(value="asd"),
                    COMMENT(value=" asd //")])

    def test_tokenizes_multi_line_comment(self):
        self.check("/* asd */",
                   [MULTI_COMMENT(value=" asd ")])

        self.check("/* /* asd */",
                   [MULTI_COMMENT(value=" /* asd ")])

        self.check("/* /* asd */",
                   [MULTI_COMMENT(value=" /* asd ")])

        self.check("/* 1 \n 2 */",
                   [MULTI_COMMENT(value=" 1 \n 2 ")])

        self.check("/* 1 \r\n 2 */",
                   [MULTI_COMMENT(value=" 1 \r\n 2 ")])

    def test_tokenizes_semi_colon(self):
        self.check("asd;",
                   [IDENTIFIER(value="asd"),
                    SEMI_COLON(value='')])

    def test_tokenizes_newline(self):
        self.check("asd\n",
                   [IDENTIFIER(value="asd"),
                    NEWLINE(value='')])

    def test_tokenizes_comma(self):
        self.check(",",
                   [COMMA(value='')])

    def test_tokenizes_parenthesis(self):
        self.check("()",
                   [LPAR(value=''),
                    RPAR(value='')])

    def test_tokenizes_hash(self):
        self.check("#",
                   [HASH(value='')])

    def test_tokenizes_equal(self):
        self.check("=",
                   [EQUAL(value='')])

    def test_escaped_newline_ignored(self):
        self.check("a\\\nb",
                   [IDENTIFIER(value='a'),
                    IDENTIFIER(value='b')])

    def test_tokenizes_keywords(self):
        self.check("module",
                   [MODULE(value='')])
        self.check("endmodule",
                   [ENDMODULE(value='')])
        self.check("package",
                   [PACKAGE(value='')])
        self.check("endpackage",
                   [ENDPACKAGE(value='')])
        self.check("parameter",
                   [PARAMETER(value='')])
        self.check("import",
                   [IMPORT(value='')])

    def test_has_location_information(self):
        self.check("`define foo", [
            PREPROCESSOR(value="define", location=(("fn.v", (0, 6)), None)),
            WHITESPACE(value=" ", location=(("fn.v", (7, 7)), None)),
            IDENTIFIER(value="foo", location=(("fn.v", (8, 10)), None)),
        ], strip_loc=False)

    def setUp(self):
        self.tokenizer = VerilogTokenizer()

    def check(self, code, tokens, strip_loc=True):
        """
        Helper method to test tokenizer
        Tokenize code and check that it matches tokens
        optionally strip location information in comparison
        """

        def preprocess(tokens):  # pylint: disable=missing-docstring
            if strip_loc:
                return [token.kind(token.value, None) for token in tokens]

            return tokens

        self.assertEqual(preprocess(list(self.tokenizer.tokenize(code, "fn.v"))),
                         tokens)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
License header sanity check
"""
from __future__ import print_function

import unittest
from os.path import join, splitext, abspath, commonprefix
from os import walk
import re
from datetime import datetime
from subprocess import Popen, PIPE, STDOUT
from vunit import ROOT
from vunit.builtins import VHDL_PATH
import vunit.ostools as ostools
from vunit.about import license_text

RE_LICENSE_NOTICE = re.compile(
    r"(?P<comment_start>#|--|//) This Source Code Form is subject to the terms of the Mozilla Public" + "\n"
    r"(?P=comment_start) License, v\. 2\.0\. If a copy of the MPL was not distributed with this file," + "\n"
    r"(?P=comment_start) You can obtain one at http://mozilla\.org/MPL/2\.0/\." + "\n"
    r"(?P=comment_start)" + "\n"
    r"(?P=comment_start) Copyright \(c\) (?P<first_year>20\d\d)(-(?P<last_year>20\d\d))?, " +
    r"Lars Asplund lars\.anders\.asplund@gmail\.com")

RE_LOG_DATE = re.compile(r'Date:\s*(?P<year>20\d\d)-\d\d-\d\d')


class TestLicense(unittest.TestCase):
    """
    Test that each file in the repository contains a valid license
    header with a correct year range.
    The correct year range is computed based on the commit history.
    """

    def test_that_a_valid_license_exists_in_source_files_and_that_global_licensing_information_is_correct(self):
        for file_name in find_licensed_files():
            code = ostools.read_file(file_name)
            self._check_license(code, file_name)
            if splitext(file_name)[1] in ('.vhd', '.vhdl', '.v', '.sv'):
                self._check_no_trailing_whitespace(code, file_name)

    def test_that_license_file_matches_vunit_license_text(self):
        with open(join(ROOT, 'LICENSE.txt'), "rU") as lic:
            self.assertEqual(lic.read(), license_text())

    def _check_license(self, code, file_name):
        """
        Check that the license header of file_name is valid
        """

        first_year, last_year = find_expected_license_years_of(file_name)
        match = RE_LICENSE_NOTICE.search(code)
        self.assertIsNotNone(match, "Failed to find license notice in %s" % file_name)
        if first_year == last_year:
            self.assertEqual(int(match.group('first_year')), first_year,
                             'Expected copyright year to be %d in %s' % (first_year, file_name))
            self.assertIsNone(match.group('last_year'), 'Expected no copyright years range in %s' % file_name)

        else:
            self.assertIsNotNone(match.group('last_year'),
                                 'Expected copyright year range %d-%d in %s' % (first_year, last_year, file_name))
            self.assertEqual(int(match.group('first_year')), first_year,
                             'Expected copyright year range to start with %d in %s' % (first_year, file_name))
            self.assertEqual(int(match.group('last_year')), last_year,
                             'Expected copyright year range to end with %d in %s' % (last_year, file_name))

    @staticmethod
    def _check_no_trailing_whitespace(code, file_name):
        """
        Check that there is no trailing whitespace within the code
        """
        for idx, line in enumerate(code.splitlines()):
            sline = line.rstrip()
            if sline == line:
                continue
            line_prefix = "%i: " % (idx + 1)
            print("Trailing whitespace violation in %s" % file_name)
            print(line_prefix + line)
            for _ in range(len(line_prefix) + len(sline)):
                print(" ", end="")
            for _ in range(len(line) - len(sline)):
                print("~", end="")
            print()
            raise AssertionError("Line %i of %s contains trailing whitespace" % (idx + 1, file_name))


def find_expected_license_years_of(file_name):
    """
    Use git history to find expected license years of file_name
    """
    proc = Popen(['git', 'log', '--follow', '--date=short', file_name],
                 bufsize=0, stdout=PIPE, stdin=PIPE, stderr=STDOUT, universal_newlines=True)
    out, _ = proc.communicate()
    first_year = None
    last_year = None
    for date in RE_LOG_DATE.finditer(out):
        first_year = int(date.group('year')) if first_year is None else min(int(date.group('year')), first_year)
        last_year = int(date.group('year')) if last_year is None else max(int(date.group('year')), last_year)

    if first_year is None and last_year is None:
        # File not in log yet, set to current year
        first_year = datetime.now().year
        last_year = first_year

    return first_year, last_year


def fix_license(file_name):
    """
    Fix license notice in file
    """
    first_year, last_year = find_expected_license_years_of(file_name)

    with open(file_name, "r") as fptr:
        text = fptr.read()

    if last_year is None or first_year == last_year:
        replacement = "Copyright (c) %i" % first_year
    else:
        replacement = "Copyright (c) %i-%i" % (first_year, last_year)

    text = re.sub(r"Copyright \(c\) (?P<first_year>20\d\d)(-(?P<last_year>20\d\d))?",
                  replacement,
                  text)

    with open(file_name, "w") as fptr:
        fptr.write(text)


def find_licensed_files():
    """
    Return all licensed files
    """
    licensed_files = []
    for root, _, files in walk(ROOT):
        for file_name in files:
            if 'preprocessed' in root:
                continue
            if 'codecs' in root:
                continue
            if root == join(ROOT, "docs"):
                continue
            if join(ROOT, ".tox") in root:
                continue
            osvvm_directory = abspath(join(VHDL_PATH, 'osvvm'))
            if is_prefix_of(osvvm_directory, abspath(join(root, file_name))):
                continue
            osvvm_integration_example_directory = abspath(
                join(ROOT, 'examples', 'vhdl', 'osvvm_integration', 'src'))
            if is_prefix_of(osvvm_integration_example_directory, abspath(join(root, file_name))):
                continue
            if splitext(file_name)[1] in ('.vhd', '.vhdl', '.py', '.v', '.sv'):
                licensed_files.append(join(root, file_name))
    return licensed_files


def is_prefix_of(prefix, of_path):
    """
    Return True if 'prefix' is a prefix of 'of_path'
    """
    return commonprefix([prefix, of_path]) == prefix


def main():
    """
    Fix license notice in all licensed files
    """
    for file_name in find_licensed_files():
        fix_license(file_name)


if __name__ == "__main__":
    main()
